''''''
import sys
import queue
import threading
import multiprocessing as mp
from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor, as_completed
import rich
from rich import progress
from . import utils



def get_pbar(pbar=None, bytes=False):
    return pbar or progress.Progress(
        "[progress.description]{task.description}",
        progress.BarColumn(),
        "[progress.percentage]{task.percentage:>3.0f}%",
        *([progress.DownloadColumn()] if bytes else [utils.MofNColumn()]),
        progress.TimeRemainingColumn(),
        progress.TimeElapsedColumn(),
        refresh_per_second=8,
    )


'''
-- Multi Process:

Bars(**kw) -> add_task(overall, **kw)
Bars.add(**kw) -> add_task(item, **kw)

RemoteBar__init__() -> --
RemoteBar.__call__(**kw) -> RemoteBar.__enter__(**kw)
RemoteBar.__enter__(**kw) -> start_task(item, **kw)
iter(RemoteBar(**kw)) -> start_task(item, **kw)
RemoteBar.update(**kw) -> update(item, **kw)

-- Single Process:

Bar(**kw) -> add_task(item, **kw)
Bar.__call__(**kw) -> iter(Bar(**kw))
Bar.update(**kw) -> update(item, **kw)

'''
    

class Bars:
    def __init__(self, desc="", pbar=None, iter=None, total=None, pool_mode='process', **kw) -> None:
        if isinstance(desc, progress.Progress):
            desc, pbar = None, desc
        self.desc = desc
        self.pbar = get_pbar(pbar)
        self._overall_kw = kw
        self.total = utils.try_len(iter, total)
        self._pq = utils.POOL_QUEUES[pool_mode](self._on_message)
        self._tasks = {}

    def __enter__(self):
        self.pbar.__enter__()
        self._pq.__enter__()
        self._tasks = {}
        # ---------------------------------------------------------------------------- #
        #                               add overall task                               #
        # ---------------------------------------------------------------------------- #
        self.overall_task = self.pbar.add_task(self.desc, total=self.total, **self._overall_kw)
        return self
   
    def __exit__(self, c,v,t):
        self._pq.__exit__(c,v,t)
        self.pbar.__exit__(c,v,t)

    def add(self, title, visible=False, **kw):
        # ---------------------------------------------------------------------------- #
        #                                   add task                                   #
        # ---------------------------------------------------------------------------- #
        task_id = self.pbar.add_task(title, visible=visible, start=False, **kw)
        self._tasks[task_id] = {}
        return RemoteBar(self._pq.queue, task_id)
    
    def close(self):
        self.__exit__(None,None,None)

    def _on_message(self, task_id, method, args, data):
        if method == 'print':
            print(*args, end='')
            # rich.print(*args, end='', sep=" ", **data)
        elif method == 'update':
            self._update(task_id, *args, **data)
        else:
            getattr(self.pbar, method)(*args, **data)

    def _update(self, task_id, **data):
        # update the task-specific progress bar
        if task_id is not None:
            self._tasks[task_id].update(**data)
            data.pop('complete', None)
            # ---------------------------------------------------------------------------- #
            #                                  update task                                 #
            # ---------------------------------------------------------------------------- #
            self.pbar.update(task_id, **data)

        # update the overall task progress bar
        n_finished = sum(bool(d and d.get('complete', False)) for d in self._tasks.values())
        # ---------------------------------------------------------------------------- #
        #                                update overall                                #
        # ---------------------------------------------------------------------------- #
        self.pbar.update(self.overall_task, completed=n_finished, total=len(self._tasks))

    @classmethod
    def ipool(cls, fn, xs, *a, n_workers=8, desc=None, mainbar_kw=None, subbar_kw=None, pool_mode='process', **kw):
        """Execute a function in a process pool with a progress bar for each task."""
        # get the arguments for each task
        items = [x if isinstance(x, args) else args(x) for x in xs]
    
        # no workers, just run the function
        if n_workers < 2:
            for i, arg in enumerate(items):
                desc_i = arg(desc or f'task {i}')
                yield arg(fn, *a, pbar=Bar, **kw)
            return

        # run the function in a process pool
        futures = []
        with utils.POOL_EXECUTORS[pool_mode](max_workers=n_workers) as executor, cls(pool_mode=pool_mode, **(mainbar_kw or {})) as pbars:
            for i, arg in enumerate(items):
                desc_i = arg(desc or f'task {i}')
                pbar = pbars.add(desc_i, **(subbar_kw or {}))
                futures.append(executor.submit(fn, *arg.a, *a, pbar=pbar, **dict(kw, **arg.kw)))
            for f in as_completed(futures):
                yield f.result()

    @classmethod
    def pool(cls, fn, xs, *a, n_workers=8, desc=None, **kw):
        return list(cls.imap(fn, xs, *a, n_workers=n_workers, desc=desc, **kw))

    # not sure which name is better
    imap = ipool
    map = pool


class args:
    def __init__(self, *a, **kw):
        self.a = a
        self.kw = kw

    def __call__(self, fn, *a, **kw):
        return fn(*self.a, *a, **dict(self.kw, **kw)) if callable(fn) else fn


class RemoteBar:
    def __init__(self, q, task_id):
        self._queue = q
        self.task_id = task_id
        self.current = 0
        self.total = 0
        self.complete = False
        self._started = False
        self.kw = {}
        self.update(0)

    @property
    def console(self):
        if getattr(self, '_console', None) is None:
            # file=_QueueFile(self._queue, self.task_id)
            self._console = rich.console.Console(file=_QueueFile(self._queue, self.task_id))
        return self._console

    def __getstate__(self):
        state = self.__dict__.copy()
        state['_console'] = None
        return state
    
    def __enter__(self, **kw):
        # ---------------------------------------------------------------------------- #
        #                                  start task                                  #
        # ---------------------------------------------------------------------------- #
        # start the task if it hasn't been started yet
        if not self._started:
            self._call('start_task', task_id=self.task_id, **kw)
            self._started = True
        return self
            
    def __exit__(self, exc_type, exc_value, tb):
        self.close()

    def close(self):
        self.update(0, complete=True)
        # ---------------------------------------------------------------------------- #
        #                                   stop task                                  #
        # ---------------------------------------------------------------------------- #
        if self._started:
            self._call('stop_task', task_id=self.task_id)
            self._started = False
        return
    
    def print(self, *a, **kw):
        self.console.print(*a, **kw)
        return self
    
    def set_description(self, desc):
        self.update(0, description=desc)
        return self

    def __call__(self, iter=None, total=None, **kw):
        # if the first argument is a string, use it as the description
        if isinstance(iter, str):
            iter, kw['description'] = None, iter
        if iter is None:
            if total is not None:
                self.total = kw['total'] = total
            # if kw or total is not None:
            #     self.update(0, **kw)
            # ---------------------------------------------------------------------------- #
            #                                  start task                                  #
            # ---------------------------------------------------------------------------- #
            self.__enter__(**kw)
            return self
        
        # --------------------------------- iterable --------------------------------- #

        self.total = kw['total'] = utils.try_len(iter, total)
        def _iter():
            # ---------------------------------------------------------------------------- #
            #                                  start task                                  #
            # ---------------------------------------------------------------------------- #
            self.__enter__(**kw)
            try:
                for x in iter:
                    yield x
                    # ---------------------------------------------------------------------------- #
                    #                                    update                                    #
                    # ---------------------------------------------------------------------------- #
                    self.update()
            finally:
                # ---------------------------------------------------------------------------- #
                #                                   stop task                                  #
                # ---------------------------------------------------------------------------- #
                self.__exit__(*sys.exc_info())
        return _iter()
    
    def _call(self, method, *args, **kw):
        self._queue.put((self.task_id, method, args, kw))

    def set(self, value):
        self.current = value
        self.update(0)
        return self

    def update(self, n=1, **kw):
        if not self._started:
            self.__enter__(**kw)

        if 'total' in kw:
            self.total = kw['total']

        # track all keyword arguments
        self.kw = dict(self.kw, **kw)
        kw = dict(self.kw)

        # calculate task progress
        self.current += n
        total = self.total if self.current or not self.total else (self.current+1)
        self.complete = total and self.current >= total
        visible = bool(total and not self.complete or not kw.get('transient', True))
        kw.setdefault('visible', visible)
        kw.setdefault('total', total)
        kw.setdefault('complete', self.complete)

        # ---------------------------------------------------------------------------- #
        #                                    update                                    #
        # ---------------------------------------------------------------------------- #
        self._call('update', completed=self.current, **kw)
        return self


class _QueueFile:
    isatty=rich.get_console().file.isatty
    def __init__(self, q, task_id):
        self._queue = q
        self.task_id = task_id
        self._buffer = []
        self.kw = {}

    def write(self, *args, **kw):
        self._buffer.extend(args)
        self.kw = kw

    def flush(self):
        if self._buffer:
            self._queue.put((self.task_id, 'print', self._buffer[:], self.kw))
            self._buffer = []
            self.kw = {}

class Bar:
    def __init__(self, desc=None, bytes=False, pbar=None, total=None, **kw):
        if isinstance(desc, progress.Progress):
            desc, pbar = None, desc
        self.pbar = get_pbar(pbar, bytes=bytes)
        # ---------------------------------------------------------------------------- #
        #                                   add task                                   #
        # ---------------------------------------------------------------------------- #
        self.task_id = self.pbar.add_task(desc, start=total is not None, total=total, **kw)
        self.pbar.__enter__()

    def __enter__(self):
        return self

    def __exit__(self, c,v,t):
        self.pbar.__exit__(c,v,t)

    def __call__(self, iter, total=None, **kw):
        with self.pbar:
            # ---------------------------------------------------------------------------- #
            #                                 update total                                 #
            # ---------------------------------------------------------------------------- #
            # set the initial total
            self.update(0, total=utils.try_len(iter, total), **kw)
            # loop through the elements
            for i, x in enumerate(iter):
                yield x
                # ---------------------------------------------------------------------------- #
                #                                    update                                    #
                # ---------------------------------------------------------------------------- #
                self.update()

    def print(self, *a, **kw):
        rich.print(*a, **kw)
        return self
    
    def set_description(self, desc):
        self.update(0, description=desc)
        return self

    def update(self, n=1, total=None, **kw):
        # start the task if it hasn't been started yet
        if total is not None:
            # ---------------------------------------------------------------------------- #
            #                                  start task                                  #
            # ---------------------------------------------------------------------------- #
            task = self.pbar._tasks[self.task_id]
            if task.start_time is None:
                self.pbar.start_task(self.task_id)
                # print('starting task', total, task, self.task_id)

        # ---------------------------------------------------------------------------- #
        #                                    update                                    #
        # ---------------------------------------------------------------------------- #
        self.pbar.update(self.task_id, advance=n, total=total, **kw)
        return self

    def close(self):
        pass


def mqdm(iter, desc=None, bytes=False, pbar=None, total=None, n_workers=0, **kw):
    return Bar(desc, bytes, pbar, total)(iter, **kw)


# ---------------------------------------------------------------------------- #
#                                   Examples                                   #
# ---------------------------------------------------------------------------- #


def example_fn(i, pbar):
    import random
    from time import sleep
    for i in pbar(range(i + 1)):
        t = random.random()*2 / 10
        sleep(t)
        pbar.print(i, "slept for", t)
        pbar.set_description("sleeping for %.2f" % t)

def my_work(n, pbar, sleep=0.2):
    import time
    for i in pbar(range(n), description=f'counting to {n}'):
        time.sleep(sleep)


def my_other_work(n, pbar, sleep=0.2):
    import time
    time.sleep(1)
    with pbar(description=f'counting to {n}', total=n):
        for i in range(n):
            pbar.update(0.5, description=f'Im counting - {n}  ')
            time.sleep(sleep/2)
            pbar.update(0.5, description=f'Im counting - {n+0.5}')
            time.sleep(sleep/2)


def example_run():
    rich.print("asdf", "asdfasdf")
    rich.print("asdf", [1,2,3], "asdfasdf")
    Bars.pool(
        example_fn, 
        range(10), 
        # desc=lambda i: f"wowow {i} :o", 
        mainbar_kw={'transient': False},
        subbar_kw={'transient': False},
        n_workers=5)

    

if __name__ == '__main__':
    import fire
    fire.Fire(example_run)


import time
import mqdm as M
from mqdm import mqdm, pool, print


# ---------------------------------------------------------------------------- #
#                                   Examples                                   #
# ---------------------------------------------------------------------------- #


def example_group():
    t0 = time.time()
    for i in range(10):
        for j in mqdm(range(100), desc=f'blah {i}'):
            time.sleep(0.005)
        print("loop")
    print(f"done in {time.time() - t0:.2f} seconds")

    t0 = time.time()
    with M.group():
        for i in range(10):
            for j in mqdm(range(100), desc=f'blah {i}'):
                time.sleep(0.005)
            print("loop")
    print(f"done in {time.time() - t0:.2f} seconds")

    t0 = time.time()
    for i in range(10):
        for j in mqdm(range(100), desc=f'blah {i}'):
            time.sleep(0.005)
        print("loop")
    print(f"done in {time.time() - t0:.2f} seconds")



def example_bar(n=8, sleep=1, transient=False, error=False, indet=False, embed=False, bp=False):
    t0 = time.time()
    xs = range(n)
    if indet:
        xs = (x for x in xs)
    for i in M.mqdm(xs, desc='example', transient=transient):
        M.set_description(f'example {i}')
        for j in M.mqdm(range(10), desc=f'blah {i}', transient=transient):
            time.sleep(0.04 * sleep)
            if j == 5 and not i % 2:
                # print("blah", i, j)
                if error: 1/0
                if embed: M.embed()
                if bp: M.bp()
    print(f"done in {time.time() - t0:.2f} seconds")


def example_fn(i, error=False, sleep=1):
    import time
    import random
    for i in mqdm(range(i + 1), desc=f'example {i}'):
        t = sleep * random.random()*2 / (i+1)
        time.sleep(t)
        print(i, "slept for", t)
        # mqdm_.set_description("sleeping for %.2f" % t)
        if error: 1/0

def example_pool(n=5, transient=False, n_workers=5, **kw):
    import time
    t0 = time.time()
    pool(
        example_fn,
        # example_bar, 
        range(n), 
        '[bold blue]Very important work',
        bar_kw={'transient': transient},
        # transient=True,
        n_workers=n_workers,
        **kw)
    print("done in", time.time() - t0, "seconds", 123)


def example_messy(n=3, transient=False, n_workers=5, **kw):
    import time
    t0 = time.time()
    example_bar(2, transient=False)
    pbar = mqdm(desc='example leftover', leave=True)
    # pbar1 = mqdm(desc='other', total=10)
    # with pbar:
    pbar.update()
    # time.sleep(1)
    pbar.update()
    # time.sleep(1)
    # pbar.__enter__()
    pbar.update()

    pool(
        example_fn, 
        range(n), 
        '[bold blue]Very important work',
        bar_kw={'transient': True},
        n_workers=n_workers,
        **kw)
    pbar.update()
    pbar.update()
    time.sleep(1)
    example_bar(3, sleep=1, transient=True)
    time.sleep(1)
    pbar.update(total=3)
    # pbar.__exit__(None, None, None)
    # time.sleep(1)
    # mqdm_.pbar.remove_task(0)
    # mqdm_.pbar.stop()
    # time.sleep(1)
    # example(2, transient=False)
    print("done in", time.time() - t0, "seconds", 123)
    pbar.close()
    # M.pbar.stop()
    # time.sleep(1)
    print("done in", time.time() - t0, "seconds", 123)
    print("done in", time.time() - t0, "seconds", 123)
    # print("done in", time.time() - t0, "seconds", 123)
    # print("done in", time.time() - t0, "seconds", 123)
    # print("done in", time.time() - t0, "seconds", 123)
    # print("done in", time.time() - t0, "seconds", 123)
    # 1/0

    # example_bar(2, transient=False)
    # M.pbar.stop()


import dataclasses
from typing import Type
from functools import wraps
import multiprocessing as mp
from multiprocessing.managers import BaseProxy, SyncManager
import multiprocessing.managers
import mqdm
import rich
from rich import progress
from .utils import T_POOL_MODE, _RedirectIO


def get_progress_cls(pool_mode: T_POOL_MODE=None):
    if pool_mode == 'process':
        return get_manager().mqdm_Progress
    return Progress

class Progress(progress.Progress):
    multiprocess = False

    def __init__(self, *columns, _tasks=None, _task_index=None, **kw):
        super().__init__(*columns, **kw)

        # save init options in case we need to recreate the object in a different process
        self._init_options = kw
        # load serialized tasks from another progress object
        if _tasks is not None:
            self._tasks = {task_id: self._load_task(**task) for task_id, task in _tasks.items()}
        self._task_index = progress.TaskID(_task_index or 0)

    def print(self, *args, **kw):
        """Print to the console."""
        rich.print(*args, **kw)

    @wraps(progress.Progress.update)
    def update(self, task_id, **kw):
        if 'description' in kw and kw['description'] is None:  # ignore None descriptions
            kw.pop('description')
        return super().update(task_id, **kw)
    
    # def stop(self):
    #     if self.live.is_started:                    
    #         self.refresh()
    #         super().stop()

    def pop_task(self, task_id, remove=None):
        """Close a task and return its serialized data."""
        try:
            data = self.dump_task(task_id)
            self.stop_task(task_id)
            if remove is None:
                remove = self._tasks[task_id].fields.get('transient', False)
            if remove:
                self.remove_task(task_id)
            return data
        except KeyError as e:
            pass

    # convert to multiprocessing proxy

    def convert_proxy(self) -> 'ProgressProxy':
        """Convert to a multiprocessing proxy object so methods can be called in another process."""
        # get current state and cleanup
        started = self.live.is_started
        tasks = self.dump_tasks()
        for task_id in tasks:
            self.remove_task(task_id)
        started or self.start()
        self.refresh()
        self.stop()

        # create proxy
        proxy = get_manager().mqdm_Progress(
            *self.columns,
            _tasks=tasks,
            _task_index=self._task_index,
            **self._init_options,
        )
        proxy.multiprocess = True
        if started:
            proxy.start()
        return proxy
    
    # def get_task_attr(self, task_id, attr, default=None):
    #     task = self._tasks[task_id]
    #     try:
    #         return getattr(task, attr)
    #     except AttributeError:
    #         return task.fields.get(attr, default)

    def dump_tasks(self):
        with self._lock:
            return {task_id: self._dump_task(task_id) for task_id in self._tasks}

    def dump_task(self, task_id):
        with self._lock:
            return self._dump_task(task_id)

    def _dump_task(self, task_id):
        task = self._tasks[task_id]
        task = {k.name: getattr(task, k.name) for k in dataclasses.fields(task) if not k.name.startswith('_')}
        return task
    
    def _load_task(self, *, start_time=None, stop_time=None, **data):
        task = progress.Task(_get_time=self.get_time, _lock=self._lock, **data)
        task.start_time = start_time
        task.stop_time = stop_time
        return task

    def load_task(self, task: dict, start=True):
        with self._lock:
            task = self._load_task(**task)
            self._tasks[task.id] = task
            if task.id >= self._task_index:
                self._task_index = progress.TaskID(task.id+1)
        if start:
            self.start_task(task.id)




def proxymethod(func):
    name = func.__name__
    @wraps(func)
    def _call(self, *a, **kw):
        return self._callmethod(name, a, kw)
    _call._is_exposed_ = True
    return _call


class ProgressProxy(BaseProxy):
    multiprocess = True
    # def __init__(self, *args, **kw):
    #     console = rich.get_console()
    #     self._redirect_io = _RedirectIO(console)
    #     self._redirect_io.enable()
    #     super().__init__(*args, **kw)

    start_task = proxymethod(Progress.start_task)
    stop_task = proxymethod(Progress.stop_task)
    add_task = proxymethod(Progress.add_task)
    remove_task = proxymethod(Progress.remove_task)
    update = proxymethod(Progress.update)
    refresh = proxymethod(Progress.refresh)
    start = proxymethod(Progress.start)
    stop = proxymethod(Progress.stop)
    print = proxymethod(Progress.print)
    # get_task_attr = proxymethod(Progress.get_task_attr)
    dump_task = proxymethod(Progress.dump_task)
    dump_tasks = proxymethod(Progress.dump_tasks)
    load_task = proxymethod(Progress.load_task)
    pop_task = proxymethod(Progress.pop_task)

ProgressProxy._exposed_ = tuple(k for k, v in ProgressProxy.__dict__.items() if getattr(v, '_is_exposed_', False))

multiprocessing.managers.ProgressProxy = ProgressProxy  # Can't pickle - attribute lookup ProgressProxy on multiprocessing.managers failed


class MqdmManager(SyncManager):
    mqdm_Progress: Type[ProgressProxy]
MqdmManager.register('mqdm_Progress', Progress, ProgressProxy)

def get_manager() -> MqdmManager:
    if getattr(mqdm, '_manager', None) is not None:
        return mqdm._manager
    mqdm._manager = manager = MqdmManager()
    manager.start()
    return manager



'''First draft implementation using manager dict.


'''

import random
from time import sleep
import threading
import multiprocessing
from concurrent.futures import ProcessPoolExecutor
from rich import progress


class Bars:
    def __init__(self, desc="[green]All jobs progress:") -> None:
        self.desc = desc
        self.pbar = progress.Progress(
            "[progress.description]{task.description}",
            progress.BarColumn(),
            "[progress.percentage]{task.percentage:>3.0f}%",
            progress.TimeRemainingColumn(),
            progress.TimeElapsedColumn(),
            refresh_per_second=8,
        )
        # Create and start a thread for reading from the queue
        reader_thread = threading.Thread(target=queue_reader, args=(queue,))
        reader_thread.start()
        self.manager = multiprocessing.Manager()

    def __enter__(self):
        self.pbar.__enter__()
        self.manager.__enter__()
        self.tasks = self.manager.dict()
        self.overall_task = self.pbar.add_task(self.desc)
        return self
   
    def __exit__(self, c,v,t):
        self.pbar.__exit__(c,v,t)
        self.manager.__exit__(c,v,t)

    def add(self, title, visible=False, **kw):
        task_id = self.pbar.add_task(title, visible=visible, start=False, **kw)
        return Bar(self.tasks, task_id)

    def update(self):
        n_finished = 0
        for task_id, data in self.tasks.items():
            data = dict(data)
            complete = data['total'] and data['completed'] >= data['total']
            visible = bool(data['total'] and not complete)
            data.setdefault('visible', visible)
            n_finished += complete
            # update the progress bar for this task:
            self.pbar.update(task_id, **data)
        self.pbar.update(self.overall_task, completed=n_finished, total=len(self.tasks))

    def as_completed(self, futures, cycle=0.1):
        while futures:
            done = [futures.pop(i) for i, f in enumerate(futures) if f.done()]
            for f in done:
                yield f
            self.update()
            sleep(cycle)

    @classmethod
    def map(cls, fn, xs, n_workers=8, desc=None, **kw):
        futures = []
        with cls() as pbars, ProcessPoolExecutor(max_workers=n_workers) as executor:
            for i, x in enumerate(xs):
                desc_i = desc(*x, **kw) if callable(desc) else desc or f'task {i}'
                futures.append(executor.submit(fn, *x, pbar=pbars.add(desc_i), **kw))
            for f in pbars.as_completed(futures):
                result = f.result()


class Bar:
    def __init__(self, tasks, task_id):
        self.tasks = tasks
        self.task_id = task_id
        self.current = 0
        self.total = 0
        self.update(0)

    def __call__(self, iter, total=None):
        try:
            self.total = len(iter)
        except TypeError:
            self.total = total
        def _iter():
            
            for x in iter:
                yield x
                self.update()
        return _iter()

    def set(self, value):
        self.current = value
        self.update(0)
        return self

    def update(self, n=1):
        self.current += n
        total = self.total if self.current or not self.total else (self.current+1)
        self.tasks[self.task_id] = {"completed": self.current, "total": total}
        return self



def example_fn(i, pbar):
    import random
    for i in pbar(range(i + 1)):
        sleep(random.random())

def pbar_test():
    Bars.map(example_fn, [[i] for i in range(10)], n_workers=5)
if __name__ == '__main__':

    import fire
    fire.Fire(pbar_test)



from contextlib import contextmanager

import rich
from rich import progress


import mqdm as M  # self
from . import proxy
from . import utils
from .utils import T_POOL_MODE
from .utils import args, executor
from .utils import embed, inp, bp, iex


_manager = None
_instances = []
_keep = False
pbar: 'proxy.Progress|proxy.ProgressProxy' = None


# ---------------------------------------------------------------------------- #
#                               Progress methods                               #
# ---------------------------------------------------------------------------- #


def _new_pbar(pool_mode: T_POOL_MODE=None, bytes=False, **kw):
    kw.setdefault('refresh_per_second', 8)
    return proxy.get_progress_cls(pool_mode)(
        "[progress.description]{task.description}",
        progress.BarColumn(bar_width=None),
        "[progress.percentage]{task.percentage:>3.0f}%",
        utils.MofNColumn(bytes=bytes),
        utils.SpeedColumn(bytes=bytes),
        utils.TimeElapsedColumn(compact=True),
        progress.TimeRemainingColumn(compact=True),
        progress.SpinnerColumn(),
        **kw,
    )


def _get_pbar(pool_mode: T_POOL_MODE=None, start=True, **kw):
    # no progress bar yet, create one
    if not M.pbar:
        # print("New progress bar", pool_mode)
        M.pbar = _new_pbar(pool_mode=pool_mode, **kw)
    # need to create multiprocess-compatible progress bar
    elif pool_mode == 'process' and not M.pbar.multiprocess:
        # print("Converting proxy")
        M.pbar = M.pbar.convert_proxy()
    if start:
        M.pbar.start()
    return M.pbar


def _clear_pbar(strict=True, force=False, soft=False):
    """Clear the progress bar."""
    if force:
        for bar in _instances[::-1]:
            bar._remove(False)
            bar.disable = True
        M.pbar.stop()
        M.pbar = None
    if M._instances:
        if strict:
            raise RuntimeError("Cannot clear progress bar while instances are still active.")
    elif not utils.is_main_process():
        if strict:
            raise RuntimeError("Cannot clear progress bar in a subprocess.")
    elif soft or M._keep:
        if M.pbar is not None:
            M.pbar.refresh()
    else:
        if M.pbar is not None:
            M.pbar.stop()
        M.pbar = None


@contextmanager
def group():
    """Group progress bars."""
    try:
        M._keep = True
        yield 
    finally:
        M._keep = False
        _clear_pbar()


# ---------------------------------------------------------------------------- #
#                            Global context methods                            #
# ---------------------------------------------------------------------------- #


def print(*args, **kw):
    """Print with rich."""
    if pbar is not None:
        return pbar.print(*args, **kw)
    return rich.print(*args, **kw)


def get(i=-1):
    """Get an mqdm instance."""
    try:
        return _instances[i]
    except IndexError:
        raise IndexError(f'No progress bar found at index {i} in list of length {len(_instances)}')


def set_description(desc, i=-1):
    """Set the description of the last progress bar."""
    return get(i).set_description(desc)


def set(i=-1, **kw):
    """Set the last progress bar."""
    return get(i).set(**kw)


def update(n=1, i=-1, **kw):
    """Update the last progress bar."""
    return get(i).update(n, **kw)


def _add_instance(bar):
    if bar not in _instances:
        _instances.append(bar)
    return bar

def _remove_instance(bar):
    while bar in _instances:
        _instances.remove(bar)


def pause(paused=True):
    """Pause the progress bars. Useful for opening an interactive shell or printing stack traces."""
    prev_paused = getattr(pbar, 'paused', False)
    if pbar is not None:
        pbar.paused = paused
        if paused:
            pbar.stop()
        else:
            pbar.start()
    return _pause_exit(prev_paused)

class _pause_exit:
    def __init__(self, prev_paused):
        self.prev_paused = prev_paused  # it was paused before we got here
        _pause_exit.last = self  # if another pause was called, ignore this one
    def __enter__(self): pass
    def __exit__(self, c, exc, t): 
        if not exc and not self.prev_paused and self is _pause_exit.last:  # dont unpause for exceptions
            pause(False)


# ---------------------------------------------------------------------------- #
#                               Primary interface                              #
# ---------------------------------------------------------------------------- #


from .bar import mqdm, pool, ipool

# more descriptive names to avoid polluting the namespace
mqpool = pool
mqipool = ipool

__all__ = [
    'mqdm',
    'mqpool',
    'mqipool',
]


''''''
import sys
import rich
from . import utils
import mqdm

import time
import random
import queue
import threading
import multiprocessing as mp

import rich
from rich import progress
from .bars import Bars



class Bars(Bars):
    def __init__(self, desc=None, *, pool_mode='process', transient=True, **kw):
        self._tasks = {}
        self._pq = POOL_QUEUES[pool_mode](self._on_message)
        self._remote = Remote(self._pq)
        self.pool_mode = pool_mode

        super().__init__(desc, transient=transient, **kw)

    def remote(self):
        return self._remote

    def __enter__(self):
        if not self._entered:
            self._pq.__enter__()
            super().__enter__()
        return self

    def __exit__(self, c,v,t):
        if self._entered:
            self._pq.__exit__(c,v,t)
            super().__exit__(c,v,t)

    def add(self, desc="", visible=False, start=False, **kw):
        return RemoteBar(self._remote, self._add_task(desc, visible=visible, start=start, **kw))

    def _add_task(self, desc="", proxy=None, **kw):
        task_id = self.pbar.add_task(description=desc or "", parent_task_id=self.task_id, **kw)
        if task_id not in self._tasks:
            self._tasks[task_id] = {}
        if proxy is not None:
            self._task_id_proxy[proxy] = task_id
        return task_id

    _task_id_proxy = {}
    def _on_message(self, task_id, method, args, data):
        if method == '__remote_add':
            if task_id in self._tasks:
                raise ValueError(f"Task with id {task_id} already exists")
            self._add_task(*args, proxy=task_id, **data)

        if isinstance(task_id, tuple):
            task_id = self._task_id_proxy[task_id]
        if method == 'raw_print':
            print(*args, end='')
        elif method == 'rich_print':
            rich.print(*args, end='', sep=" ", **data)
        elif method == 'update':
            self._update(task_id, *args, **data)
        elif method == 'start_task':
            self._tasks[task_id]['complete'] = False
            self.pbar.start_task(*args, task_id=task_id, **data)
        elif method == 'stop_task':
            self._tasks[task_id]['complete'] = True
            self.pbar.stop_task(*args, task_id=task_id, **data)
            self._update(None)
        elif method == '__pause':
            mqdm.pause(*args, **data)
        else:
            getattr(self.pbar, method)(*args, **data)

    def _update(self, task_id, **data):
        if task_id is not None:
            # -------------------------------- update task ------------------------------- #
            # update the task-specific progress bar
            self.pbar.update(task_id, **data, refresh=False)

            # update progress bar visibility
            task = self.pbar._tasks[task_id]
            current = task.completed
            total = task.total
            transient = task.fields.get('transient', True)
            complete = total is not None and current >= total
            task.visible = bool(total is not None and not complete or not transient)
            self._tasks[task_id]['complete'] = complete

        # ------------------------------ update overall ------------------------------ #
        n_finished = sum(bool(d.get('complete', False)) for d in self._tasks.values())
        self.pbar.update(self.task_id, completed=n_finished, total=len(self))



class Remote:
    _console = None
    def __init__(self, queue):
        self.queue = queue

    def _new(self):
        return RemoteBar(self.queue, self.queue.random_id())
    
    def _call(self, task_id, method, *args, **kw):
        self.queue.put((task_id, method, args, kw))

    def __call__(self, **kw):
        bar = self._new()
        self._call(bar.task_id, '__remote_add', (), bar._process_args(**kw))
        return bar

    @property
    def console(self):
        if self._console is None:
            self._console = rich.console.Console(file=utils.QueueFile(self.queue))
        return self._console

    def __getstate__(self):
        state = self.__dict__.copy()
        state['_console'] = None
        return state
    
    def __setstate__(self, state):
        self.__dict__.update(state)
        mqdm._remote = self
    
    # -------------------------- Top Level mqdm Aliases -------------------------- #
    
    def pause(self, paused=True):
        self._call(None, '__pause', paused)

    def print(self, *a, **kw):
        self.console.print(*a, **kw)
        return self
    
    def get(self, i=-1):
        return mqdm.get(i)

    def set_description(self, desc, i=-1):
        return mqdm.set_description(desc)


class RemoteBar:
    _entered = False
    _get_desc = None
    total = None
    def __init__(self, remote, task_id):
        self._remote = remote
        self.task_id = task_id

    def _call(self, method, *args, **kw):
        self._remote._call(self.task_id, method, *args, **kw)

    def __setstate__(self, state):
        self.__dict__.update(state)
        mqdm._add_instance(self)

    def __enter__(self, **kw):
        if not self._entered:
            self._call('start_task', **kw)
            mqdm._add_instance(self)
            self._entered = True
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if self._entered:
            self._call('stop_task')
            mqdm._remove_instance(self)
            self._entered = False

    def __del__(self):
        try:
            if sys.meta_path is None:
                return 
            self.__exit__(None, None, None)
        except (AttributeError, ImportError, BrokenPipeError, FileNotFoundError) as e:
            pass

    def __len__(self):
        return self.total or 0

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._iter)
    
    def _get_iter(self, iter, **kw):
        for i, x in enumerate(iter):
            self.update(i>0, arg_=x)
            yield x
        self.update()

    def __call__(self, iter=None, total=None, desc=None, **kw):
        if isinstance(iter, str) and desc is None:  # infer string as description
            iter, kw['description'] = None, iter
        if iter is None:
            return self.update(total=total, **kw)

        total = utils.try_len(iter, self.total) if total is None else total
        self.update(0, total=total, description=desc or ..., **kw)
        def _with_iter():
            if self._entered:
                yield from self._get_iter(iter, **kw)
                return
            with self:
                yield from self._get_iter(iter, **kw)
        self._iter = _with_iter()
        return self

    def print(self, *a, **kw):
        self._remote.print(*a, **kw)
        return self

    def set_description(self, desc):
        return self.update(0, description=desc or "")

    def _process_args(self, *, arg_=..., **kw):
        kw = {k: v for k, v in kw.items() if v is not ...}
        if 'total' in kw:
            self.total = kw['total']

        # get description
        if 'desc' in kw:
            kw['description'] = kw.pop('desc')
        if 'description' in kw and callable(kw['description']):
            self._get_desc = kw.pop('description')
        if 'description' not in kw and self._get_desc is not None and arg_ is not ...:
            kw['description'] = self._get_desc(arg_)
        if 'description' in kw and kw.get('description') is None:
            kw['description'] = ''

        return kw

    def update(self, n=1, *, arg_=..., **kw):
        kw = self._process_args(arg_=arg_, **kw)
        if n or kw:
            self._call('update', advance=n, **kw)
        return self






class QueueFile:
    isatty=rich.get_console().file.isatty
    def __init__(self, q, task_id=None):
        self._queue = q
        self.task_id = task_id
        self._buffer = []
        self.kw = {}

    def write(self, *args, **kw):
        self._buffer.extend(args)
        self.kw = kw

    def flush(self):
        if self._buffer:
            self._buffer, buffer = [], self._buffer
            self._queue.put((self.task_id, 'raw_print', buffer, self.kw))
            self.kw = {}


class BaseQueue:
    def __init__(self, fn):
        self._fn = fn

    def __enter__(self):
        return self

    def __exit__(self, c,v,t):
        pass

    def random_id(self):
        return (0, id(self), time.time(), random.random())


class SequentialQueue(BaseQueue):
    '''An event queue to respond to events in the main thread.'''
    def __init__(self, fn):
        self._fn = fn
        self.queue = self

    def put(self, xs):
        self._fn(*xs)


class MsgQueue(BaseQueue):
    _max_cleanup_attempts = 100
    def __init__(self, fn):
        self._fn = fn
        self._closed = False
        self._thread = threading.Thread(target=self._monitor, daemon=True)

    def __getstate__(self):
        state = self.__dict__.copy()
        state['_thread'] = None
        state['_fn'] = None
        return state

    def __enter__(self):
        if self._thread is None: return self
        self._closed = False
        try:
            self._thread.start()
        except RuntimeError:
            pass
        return self
    
    def __exit__(self, c,v,t):
        self._closed = True
        if self._thread is None: return
        self._thread.join()
        for i in range(self._max_cleanup_attempts):
            if not self._read(timeout=0.005):
                break

    def _read(self, timeout=0.1):
        try:
            xs = self.queue.get(timeout=timeout)
        except queue.Empty:
            return False
        self._fn(*xs)
        return True

    def _monitor(self):
        while not self._closed:
            self._read()

    def put(self, xs, **kw):
        self.queue.put(xs, **kw)

    def get(self, xs, **kw):
        return self.queue.get(xs, **kw)

    def raise_exception(self, e):
        pass


class ThreadQueue(MsgQueue):
    '''An event queue to respond to events in a separate thread.'''
    def __init__(self, fn):
        super().__init__(fn)
        self.queue = queue.Queue()
    #     self._exc_info = None

    def random_id(self):
        return (threading.get_ident(), random.random(), id(self), time.time(), random.random())

    # def __enter__(self):
    #     self._exc_info = None
    #     return super().__enter__()

    # def raise_exception(self):
    #     if self._exc_info:
    #         raise self._exc_info[1].with_traceback(self._exc_info[2])

    # def _read(self, timeout=0.1):
    #     try:
    #         xs = self.queue.get(timeout=timeout)
    #     except queue.Empty:
    #         return False
    #     try:
    #         self._fn(*xs)
    #     except Exception as e:
    #         self._exc_info = sys.exc_info()
    #     return True


class ProcessQueue(MsgQueue):
    '''An event queue to respond to events in a separate process.'''
    def __init__(self, fn, manager=None):
        self._self_managed = manager is not None
        self._manager = manager or mp.Manager()
        # self._manager = manager or get_manager() #mp.Manager()
        self.queue = self._manager.Queue()
        super().__init__(fn)

    def random_id(self):
        return (mp.current_process().pid, id(self), time.time(), random.random())

    def __getstate__(self):
        state = self.__dict__.copy()
        state['_manager'] = None
        state['_thread'] = None
        state['_fn'] = None
        return state

    def __enter__(self):
        if self._thread is None: return self
        if self._self_managed:
            self._manager.__enter__()
        super().__enter__()
        return self
    
    def __exit__(self, c,v,t):
        if self._thread is None: return
        super().__exit__(c,v,t)
        if self._self_managed:
            self._manager.__exit__(c,v,t)


POOL_QUEUES = {
    'thread': ThreadQueue,
    'process': ProcessQueue,
    'sequential': SequentialQueue,
    None: SequentialQueue,
}



import sys
from typing import Literal, cast, TextIO, IO, Optional
from concurrent.futures import Future, Executor, ProcessPoolExecutor, ThreadPoolExecutor
from concurrent.futures._base import FINISHED, RUNNING
import multiprocessing as mp
import rich
from rich import progress
from rich.prompt import Prompt
from rich.console import Text

import mqdm as M


# ---------------------------------------------------------------------------- #
#                                     Utils                                    #
# ---------------------------------------------------------------------------- #


def is_main_process():
    """Check if the current process is the main process."""
    return mp.current_process().name == 'MainProcess'


class args:
    '''Storing Function Arguments for later.
    
    Example:
    ```
    def fn(a, b=2, c=3):
        print(a, b, c)

    fn_args = [args(i, c=i*2) for i in range(3)]
    for arg in fn_args:
        arg(fn, b=2)
    ```
    '''
    def __init__(self, *a, **kw):
        self.a = a
        self.kw = kw

    def __getitem__(self, i):
        return self.a[i] if isinstance(i, int) else self.kw[i]

    def __call__(self, fn, *a, **kw):
        return fn(*self.a, *a, **dict(self.kw, **kw)) if callable(fn) else fn
    
    @classmethod
    def call(cls, fn, x, *a, **kw):
        return cls.from_item(x)(fn, *a, **kw)
    
    @classmethod
    def from_item(cls, x, *a, **kw):
        return x.merge_general(*a, **kw) if isinstance(x, cls) else cls(x, *a, **kw)

    @classmethod
    def from_items(cls, items, *a, **kw):
        return [cls.from_item(x, *a, **kw) for x in items]

    def merge_general(self, *a, **kw):
        return args(*self.a, *a, **dict(kw, **self.kw))


def maybe_call(fn, *a, **kw):
    """Call the value if it is callable. Otherwise, return it."""
    return fn(*a, **kw) if callable(fn) else fn


def try_len(it, default=None):
    """Try to get the length of an object, returning a default value if it fails."""
    if it is None:
        return default
    if isinstance(it, int):
        return it
    try:
        return len(it)
    except TypeError:
        pass

    try:
        x = type(it).__length_hint__(it)
        return x if isinstance(x, int) else default
    except (AttributeError, TypeError):
        return default

# ---------------------------------------------------------------------------- #
#                         Concurrent Futures Executors                         #
# ---------------------------------------------------------------------------- #

class SequentialFuture(Future):
    def __init__(self, fn, *args, **kwargs):
        super().__init__()
        self._fn = fn
        self._args = args
        self._kwargs = kwargs
        self._evaluated = False
        with self._condition:  # so as_completed will return it
            self._state = FINISHED

    def _evaluate(self):
        if not self._evaluated:
            with self._condition:
                self._state = RUNNING
            try:
                self.set_result(self._fn(*self._args, **self._kwargs))
            except Exception as exc:
                self.set_exception(exc)
            self._evaluated = True

    def result(self, timeout=None):
        self._evaluate()
        return super().result(timeout)
    
    def exception(self, timeout=None):
        self._evaluate()
        return super().exception(timeout)


class SequentialExecutor(Executor):
    def __init__(self, max_workers=None, initializer=None, initargs=()):
        super().__init__()
        self._initializer = initializer
        self._initargs = initargs

    def __enter__(self):
        if self._initializer is not None:
            self._initializer(*self._initargs)
        return super().__enter__()

    def submit(self, fn, *args, **kwargs):
        return SequentialFuture(fn, *args, **kwargs)


import threading
_thread_local_data = threading.local()
def pbar_initializer(pbar, defaults=None):
    """Initialize the progress bar for the worker thread/process."""
    if pbar is not None:
        M.pbar = pbar
    _thread_local_data.defaults = defaults or {}

def _get_local(key, default=None):
    """Get a thread-local variable."""
    return getattr(_thread_local_data, key, default)

# def _pbar_initializer(pbar, parent_task_id):
#     """Initialize the progress bar for the worker thread/process."""
#     M.pbar = pbar
#     _thread_local_data.parent_task_id = parent_task_id

T_POOL_MODE = Literal['process', 'thread', 'sequential', None]
POOL_EXECUTORS = {
    'thread': ThreadPoolExecutor,
    'process': ProcessPoolExecutor,
    'sequential': SequentialExecutor,
    None: SequentialExecutor,
}

def executor(pool_mode: T_POOL_MODE='process', bar_kw: dict=None, **kw) -> Executor:
    """Return the appropriate executor for the pool mode of the progress bar."""
    pbar = M._get_pbar(pool_mode=pool_mode)
    return POOL_EXECUTORS[pool_mode](initializer=pbar_initializer, initargs=[pbar, bar_kw or {}], **kw)


# ---------------------------------------------------------------------------- #
#                                  Debug Tools                                 #
# ---------------------------------------------------------------------------- #


def embed(*a, prompt='ipython?> ', exit_prompt=True):
    """Embed an IPython shell in the current environment. This will make sure the progress bars don't interfere.
    
    This function is useful for debugging and interactive exploration.

    Does not work in subprocesses for obvious reasons.

    .. code-block:: python

        import mqdm

        for i in mqdm_.mqdm(range(10)):
            if i == 5:
                mqdm_.embed()
    """
    with M.pause():
        from ._embed import embed
        if not prompt or _Prompt.ask(Text(f'{prompt}', style="dim cyan")): 
            a and M.print(*a)
            embed(colors='neutral', stack_depth=1)
            exit_prompt and _Prompt.ask(Text('continue?> ', style="bold magenta"))


class _Prompt(Prompt):
    prompt_suffix = '\033[F'


def inp(prompt=''):
    """Prompt for input in the terminal. This function is useful for debugging and interactive exploration."""
    with M.pause():
        return _Prompt.ask(Text(prompt or '', style="dim cyan"))


def bp(*a, prompt='ipython?> '):
    """Breakpoint"""
    with M.pause():
        if not prompt or _Prompt.ask(Text(prompt, style="dim cyan")):
            a and M.print(*a)
            breakpoint()


def iex(func):
    """Decorator to embed an IPython shell in the current environment when an exception is raised. This makes sure the progress bars don't interfere.
    
    This lets you do post-mortem debugging of the Exception stack trace.

    Does not work in subprocesses for obvious reasons.
    """
    import functools, fnmatch
    from pdbr import pdbr_context
    # from ipdb import iex
    @pdbr_context()
    def inner(*a, **kw):
        _rich_traceback_omit = True
        try:
            return func(*a, **kw)
        except:
            M.pause(True)
            rich.console.Console().print_exception(suppress=[m for k, m in sys.modules.items() if any(
                fnmatch.fnmatch(k, p) for p in ['fire', 'concurrent.futures', 'threading', 'multiprocessing'])])
            cmds='h: help, u: up, d: down, l: code, v: vars, vt: varstree, w: stack, i {var}: inspect'
            rich.print("\n[bold dim]Commands - [/bold dim] " + ", ".join("[bold green]{}[/bold green]:[dim]{}[/dim]".format(*c.split(':')) for c in cmds.split(', ')))
            raise
    @functools.wraps(func)
    def outer(*a, **kw):
        try:
            return inner(*a, **kw)
        finally:
            M.pause(False)
    return outer


# ---------------------------------------------------------------------------- #
#                         Custom Progress Column Types                         #
# ---------------------------------------------------------------------------- #


class MofNColumn(progress.DownloadColumn):
    '''A progress column that shows the current vs. total count of items.'''
    def __init__(self, bytes=False, separator="/", **kw):
        self.bytes = bytes
        self.separator = separator
        super().__init__(**kw)

    def render(self, task):
        if self.bytes:
            return super().render(task)
        total = f'{int(task.total):,}' if task.total is not None else "?"
        return progress.Text(
            f"{int(task.completed):,d}{self.separator}{total}",
            style="progress.download",
            justify='right'
        )


class SpeedColumn(progress.TransferSpeedColumn):
    """Renders human readable transfer speed."""
    def __init__(self, bytes=False, unit_scale=1, **kw):
        self.bytes = bytes
        self.unit_scale = unit_scale
        super().__init__(**kw)

    def render(self, task):
        """Show data transfer speed."""
        if self.bytes:
            return super().render(task)
        speed = task.finished_speed or task.speed
        if speed is None:
            return progress.Text("", style="progress.data.speed")
        end = 'x/s'
        if speed < 1:
            speed = 1 / speed
            end = 's/x'
        unit, suffix = progress.filesize.pick_unit_and_suffix(
            int(speed), ["", "×10³", "×10⁶", "×10⁹", "×10¹²"], 1000)
        return progress.Text(f"{speed/unit:.1f}{suffix}{end}", justify='right', style="progress.data.speed")


class TimeElapsedColumn(progress.TimeRemainingColumn):
    """Renders time elapsed."""
    def __init__(self, compact: bool = False, **kw):
        self._compact = compact
        super().__init__(**kw)

    def render(self, task):
        """Show time elapsed."""
        elapsed = task.finished_time if task.finished else task.elapsed
        if elapsed is None:
            return progress.Text("--:--" if self._compact else "-:--:--", style="progress.elapsed")
        minutes, seconds = divmod(int(elapsed), 60)
        hours, minutes = divmod(minutes, 60)
        return progress.Text(
            (f"{hours:d}:" if hours or not self._compact else "") + 
            f"{minutes:02d}:{seconds:02d}", style="progress.elapsed")

# class LogBarColumn(progress.BarColumn):
#     def render(self, task):
#         return progress.Group(
#             super().render(task),
#             progress.Text(task.description, style="progress.description"),
#         )


from rich.file_proxy import FileProxy
class _RedirectIO:
    def __init__(self, console: rich.console.Console, redirect_stdout: bool = True, redirect_stderr: bool = True):
        self.console_compatible = console.is_terminal or console.is_jupyter
        self._redirect_stdout = redirect_stdout
        self._redirect_stderr = redirect_stderr
        self._restore_stdout: Optional[IO[str]] = None
        self._restore_stderr: Optional[IO[str]] = None

    def __enter__(self):
        self.enable()
        return self
    
    def __exit__(self, *exc):
        self.disable()

    def __del__(self):
        self.disable()

    def enable(self) -> None:
        """Enable redirecting of stdout / stderr."""
        if self.console_compatible:
            if self._redirect_stdout and not isinstance(sys.stdout, FileProxy):
                self._restore_stdout = sys.stdout
                sys.stdout = cast("TextIO", FileProxy(self.console, sys.stdout))
            if self._redirect_stderr and not isinstance(sys.stderr, FileProxy):
                self._restore_stderr = sys.stderr
                sys.stderr = cast("TextIO", FileProxy(self.console, sys.stderr))

    def disable(self) -> None:
        """Disable redirecting of stdout / stderr."""
        if self._restore_stdout:
            sys.stdout = cast("TextIO", self._restore_stdout)
            self._restore_stdout = None
        if self._restore_stderr:
            sys.stderr = cast("TextIO", self._restore_stderr)
            self._restore_stderr = None


import setuptools

setuptools.setup(
    name='mqdm',
    version='1.1.0',
    description='cross-process progress bars',
    long_description=open('README.md').read().strip(),
    long_description_content_type='text/markdown',
    packages=setuptools.find_packages(),
    install_requires=[
        'rich', 'pdbr'
    ],
    extras_require={
        'test': [
            'fire'
        ]
    })


''''''
from functools import wraps
from concurrent.futures import as_completed, Executor
from typing import Callable, Iterable, Literal
from . import utils
from .bar import mqdm
import mqdm

T_POOL_MODE = Literal['process', 'thread', 'sequential', None]

class Bars(mqdm):
    """A progress bar that contains multiple sub progress bars.

    The main progress bar is used to track the completion of the sub progress bars.

    TODO: The main progress bar's estimated time should incorporate the estimated times of the sub progress bars.
    """
    _get_sub_desc = None
    def __init__(self, desc: str|Callable=None, *, pool_mode: T_POOL_MODE='process', transient=True, bar_kw=None, **kw):
        kw = {**(bar_kw or {}), 'transient': transient, **kw}
        super().__init__(desc, pool_mode=pool_mode, **kw)
        # default child progress bar keyword arguments
        self._subbar_kw = {
            'transient': transient, 
            'disable': self.disabled, 
            'visible': False, 
            'start': False,
            **(bar_kw or {}),
        }

    def __getstate__(self):
        state = super().__getstate__()
        state['_get_sub_desc'] = None  # cannot pickle lambda
        return state

    def _get_iter(self, iter, **kw):
        for i, x in enumerate(iter):
            # for each item, create a child progress bar
            pbar = self.add(desc=utils.maybe_call(self._get_sub_desc, x, i) or '', **kw)
            yield x, pbar

    # def __call__(self, iter, desc=None, DESC=None, **kw):
    #     """Iterate over an iterable with a progress bar."""
    #     if desc is None:
    #         self._get_sub_desc = lambda x, i: f'task {i}'
    #     elif isinstance(desc, str):
    #         self._get_sub_desc = lambda x, i: desc.format(x=x, i=i)
    #     else:
    #         self._get_sub_desc = desc
    #     return super().__call__(iter, DESC, **kw)

    def _process_args(self, *, task_desc=None, **kw):
        if task_desc is not None:
            self._get_sub_desc = task_desc
        return super()._process_args(**kw)

    def add(self, desc="", **kw):
        """Add a sub progress bar to the main progress bar."""
        return mqdm(desc, parent_task_id=self.task_id, **{**self._subbar_kw, **kw, 'pool_mode': self.pool_mode})

    @classmethod
    def mqdms(cls, iter=None, desc=None, DESC=None, bytes=False, transient=False, bar_kw={}, **kw):
        """Create a progress bar group for each item in an iterable.
        
        Args:
            iter (Iterable): The iterable to create subtasks for.
            desc (str, Callable, optional): The description of the sub progress bars. Can be a function that takes the item and index as arguments. Defaults to (lambda x, i: f'task {i}').
            main_desc (str, optional): The description of the main progress bar. 
            bytes (bool, optional): Whether to display the bytes processed. Defaults to False.
            transient (bool, optional): Whether to remove the progress bar after completion. Defaults to False.
            subbar_kw (dict, optional): Additional keyword arguments for the sub progress bars.
            **kw: Additional keyword arguments for the main progress bar.
        """
        return cls(desc=DESC, bytes=bytes, transient=transient, bar_kw=bar_kw)(iter, desc, **kw)

    @classmethod
    def ipool(
            cls, 
            fn: Callable, 
            iter: Iterable, 
            desc: str='', 
            task_desc: str|Callable=None,
            main_kw: dict={}, 
            bar_kw: dict={}, 
            n_workers: int=8, 
            pool_mode: T_POOL_MODE='process', 
            ordered_: bool=False, 
            squeeze_: bool=False,
            results_: list|None=None,
            **kw) -> Iterable:
        """Execute a function in a process pool with a progress bar for each task.
        
        Args:
            fn (Callable): The function to execute.
            iter (Iterable): The iterable to iterate over.
            desc (str, optional): The description of the main progress bar. 
            main_kw (dict, optional): Additional keyword arguments for the main progress bar.
            bar_kw (dict, optional): Additional keyword arguments for the sub progress bars.
            n_workers (int, optional): The number of workers in the pool. Defaults to 8.
            pool_mode (str, optional): The mode of the pool. Can be 'process', 'thread', 'sequential'. Defaults to 'process'.
            ordered_ (bool, optional): Whether to yield the results in order. Defaults to False.
            squeeze_ (bool, optional): Whether to skip the pool and main progress bar if there is only one item in the iterable. Defaults to False.
            results_ (list, optional): The list to append the results to. Useful to collect partial results in case of exception.
        """
        # no workers, just run sequentially
        if n_workers in {0, 1}:
            pool_mode = None

        # if the iterable is a single item, just run the function
        if squeeze_ and utils.try_len(iter, -1) == 1:
            arg = utils.args.from_item(iter[0], **kw)
            yield arg(fn, mqdm=mqdm.mqdm)
            if results_ is not None:
                results_.append(x)
            return

        # initialize progress bars
        pbars = cls.mqdms(
            iter, 
            desc=desc,
            task_desc=task_desc or (lambda x, i: f'task {i}'),
            pool_mode=pool_mode, 
            bar_kw=bar_kw, 
            **(main_kw or {})
        )

        try:
            with pbars:
                # no multiple workers, just run the function sequentially
                if pool_mode in {'sequential', None}:
                    for arg, pbar in pbars:
                        arg = utils.args.from_item(arg, **kw)
                        x = arg(fn, pbar=pbar)
                        if results_ is not None:
                            results_.append(x)
                        yield x
                    return

                # run the function in a process/thread pool
                with pbars.executor(max_workers=n_workers) as executor:
                    futures = []
                    for arg, pbar in pbars:
                        arg = utils.args.from_item(arg, **kw)
                        futures.append(executor.submit(fn, *arg.a, pbar=pbar, **arg.kw))

                    # get function results
                    for f in futures if ordered_ else as_completed(futures):
                        x = f.result()
                        if results_ is not None:
                            results_.append(x)
                        yield x
        except Exception as e:
            # pause the progress bar so it doesn't interfere with the traceback
            pbars.remove_task()
            mqdm.pause()
            raise

    @classmethod
    @wraps(ipool, ['__annotations__', '__doc__', '__type_params__'])
    def pool(cls, *a, **kw):
        return list(cls.ipool(*a, **kw))

    def executor(self, **kw) -> Executor:
        """Return the appropriate executor for the pool mode of the progress bar."""
        return utils.POOL_EXECUTORS[self.pool_mode](initializer=mqdm.proxy.pbar_initializer, initargs=[mqdm.pbar, self.task_id], **kw)

# ---------------------------------------------------------------------------- #
#                                   Examples                                   #
# ---------------------------------------------------------------------------- #


def example_fn(i, mqdm, error=False, sleep=1):
    import time
    import random
    for i in mqdm(range(i + 1)):
        t = sleep * random.random()*2 / (i+1)
        time.sleep(t)
        mqdm.print(i, "slept for", t)
        # mqdm.set_description("sleeping for %.2f" % t)
        if error: 1/0


def my_work(n, mqdm, sleep=0.2):
    import time
    for i in mqdm(range(n), description=f'counting to {n}'):
        time.sleep(sleep)


def my_other_work(n, mqdm, sleep=0.2):
    import time
    time.sleep(1)
    with mqdm(description=f'counting to {n}', total=n) as pbar:
        for i in range(n):
            mqdm.update(0.5, description=f'Im counting - {n}  ')
            time.sleep(sleep/2)
            mqdm.update(0.5, description=f'Im counting - {n+0.5}')
            time.sleep(sleep/2)
            mqdm.set_description(f'AAAAA - {n+1}')


def example(n=10, transient=False, n_workers=5, **kw):
    import time
    t0 = time.time()
    mqdm.pool(
        example_fn, 
        # my_work, 
        # my_other_work,
        range(n), 
        '[bold blue]Very important work',
        # main_kw={'transient': transient},
        bar_kw={'transient': transient},
        n_workers=n_workers,
        **kw)
    mqdm.print("done in", time.time() - t0, "seconds", 123)



''''''
import sys
from functools import wraps
from typing import Callable, Iterable
from concurrent.futures import as_completed

from . import print, T_POOL_MODE
from . import utils
import mqdm as M



class mqdm:
    """Create a progress bar for an iterable. (tqdm-like interface)
        
        Args:
            iter (Iterable): The iterable to iterate over.
            desc (str): The description of the progress bar.
            bytes (bool): Whether to show bytes transferred.
            pbar (rich.progress.Progress): An existing progress bar to use.
            transient (bool): Whether to remove the progress bar after completion.
            disable (bool): Whether to disable the progress bar.
            **kw: Additional keyword arguments to pass to the progress bar.
        """
    task_id = None           # the task ID of the progress bar
    pbar = None              # the progress bar instance
    _desc = None             # the description of the progress bar
    _total = None            # the total number of items to iterate over
    _n = 0                   # the number of items completed
    _iter = None             # the item iterator
    _items = None            # the items of the iterator (for indexing)
    _entered = False         # whether the progress bar has called __enter__()
    _started = False         # whether the progress bar has beed started (for lazy start)
    _task_dict = None        # the dumped task
    get_desc = None          # a function to get the description

    def __init__(self, iter=None, desc=None, *, disable=False, task_id=None, pool_mode=None, progress_kw=None, init_kw=None, **kw):
        if isinstance(iter, str) and desc is None:  # infer string as description
            iter, desc = None, iter

        self.disable = disable

        # initialize progress bar task
        init_kw = {**utils._get_local('defaults', {}), **kw, **(init_kw or {})}
        init_kw = self._process_args(description=desc, **init_kw)
        if not self.disable:
            pbar = M._get_pbar(pool_mode=pool_mode, **(progress_kw or {}))
            M._add_instance(self)
            if task_id is None:
                init_kw.setdefault('start', True) # self._total is not None
                init_kw.setdefault('total', self._total)
                init_kw.setdefault('description', '')
                task_id = M.pbar.add_task(**init_kw)
            else:
                kw = {**init_kw, **kw}
        self.task_id = task_id
        self._started = True

        # if we have an iterable, start the progress bar
        self(iter, desc=..., **kw)

    # ------------------------------ Dunder methods ------------------------------ #

    def __repr__(self) -> str:
        return f'<mqdm[{self.task_id}]({self._n}/{self._total}, {self._desc or ""!r})>'

    def __getstate__(self):
        state = self.__dict__.copy()
        state['get_desc'] = None  # cannot pickle lambda functions
        state['_iter_items'] = None  # cannot pickle iterators
        state['_iter'] = None  # cannot pickle iterators
        return state

    def __getitem__(self, i: int):
        """Get an item at index."""
        return self._items[i]
    
    # def __bool__(self) -> bool:
    #     return self._total is None or self._total

    def __len__(self) -> int:
        """Get the total number of items."""
        return self._total or 0

    def __iter__(self) -> 'mqdm':
        return self

    def __next__(self):
        return next(self._iter)
    
    # ----------------------------- Lifecycle methods ---------------------------- #
    
    def __enter__(self) -> 'mqdm':
        self._entered = True
        return self.open()

    def __exit__(self, c,v,t):
        self._entered = False
        self.close()

    def __del__(self):
        try:
            if sys.meta_path is None:
                return 
            self.close()
        except (ImportError, BrokenPipeError, FileNotFoundError) as e:
            pass

    # ----------------------------- Iteration methods ---------------------------- #

    def _get_iter(self, iter, **kw):
        for i, x in enumerate(iter):
            self.update(i>0, arg_=x, i_=i)
            yield x
        self.update(1)

    def __call__(self, iter, desc=None, total=None, **kw) -> 'mqdm':
        """Iterate over an iterable with a progress bar."""
        if isinstance(iter, str) and desc is None:  # infer string as description
            iter, desc = None, iter
        if iter is None:  # no iterable yet
            return self.update(0, total=total, **kw)
        if isinstance(iter, int):  # implicit range
            iter = range(iter)

        total = utils.try_len(iter, self._total) if total is None else total
        self.update(0, total=total, description=desc, **kw)
        def _with_iter():
            if self._entered:  # already called __enter__, don't call __exit__() when leaving
                yield from self._get_iter(iter, **kw)
                return
            with self:  # call __enter__() and __exit__()
                yield from self._get_iter(iter, **kw)
        self._items = iter
        self._iter = _with_iter()
        return self
    
    # ------------------------------ Internal methods ----------------------------- #

    def _attach(self):
        """Attach the task to the progress bar."""
        if self.disable: return

        M._get_pbar()
        M._add_instance(self)
        if self._task_dict is not None:
            M.pbar.load_task(self._task_dict)
            M.pbar.start_task(self.task_id)
            self._task_dict = None
        self.set(total=self._total)

    def _detach(self, remove=None, soft=False):
        """Detach the task from the progress bar."""
        if self.disable or M.pbar is None: return

        # stop and remove task
        if self._task_dict is None:
            self._task_dict = M.pbar.pop_task(self.task_id, remove=remove)
        M._remove_instance(self)
        M._clear_pbar(strict=False, soft=soft)

    def _process_args(self, *, append_total=None, arg_=..., i_: int=None, **kw) -> dict:
        """Process keyword arguments for updates."""
        kw = {k: v for k, v in kw.items() if v is not ...}
        if 'leave' in kw:  # tqdm compatibility
            kw['transient'] = not kw.pop('leave')

        # get progress and total
        if 'total' in kw:  # keep local copy of total
            self._total = kw['total']
        if append_total:
            kw['total'] = self._total = (self._total or 0) + append_total
        if kw.get('advance'):
            kw['advance'] = int(kw['advance'])
            self._n += kw['advance']
        if kw.get('advance') == 0:
            kw.pop('advance')
        if kw.get('completed') is not None:
            self._n = kw['completed'] = int(kw['completed'])

        # get description
        if 'desc' in kw:  # tqdm compatibility
            kw['description'] = kw.pop('desc')
        if 'description' in kw and callable(kw['description']):  # handle dynamic descriptions
            self.get_desc = kw.pop('description')
        if kw.get('description') is None and self.get_desc is not None and arg_ is not ...:
            kw['description'] = self.get_desc(arg_, i_)
        if kw.get('description') is not None:
            self._desc = kw['description']

        return kw

    # --------------------------- tqdm-like attributes --------------------------- #

    @property
    def n(self) -> int:
        """The number of items completed."""
        return self._n
    
    @n.setter
    def n(self, n: int):
        """Set the number of items completed."""
        self.set(completed=n)

    @property
    def total(self) -> int:
        """The total number of items to iterate over."""
        return self._total
    
    @total.setter
    def total(self, total: int):
        """Set the total number of items to iterate over."""
        self.set(total=total)

    # ------------------------------ public methods ------------------------------ #

    def open(self):
        """Add the task to the progress bar."""
        self._attach()
        return self

    def close(self, remove=None):
        """Remove the task from the progress bar."""
        self._detach(remove=remove)
        return self

    def print(self, *a, **kw) -> 'mqdm':
        """Print above the progress bar."""
        M.print(*a, **kw)
        return self

    def set_description(self, desc) -> 'mqdm':
        """Set the description of the progress bar."""
        return self.update(0, description=desc)

    def update(self, advance: int=1, **kw) -> 'mqdm':
        """Increment the progress bar."""
        return self.set(advance=advance, **kw)

    def set(self, **kw) -> 'mqdm':
        """Update progress bar fields."""
        kw = self._process_args(**kw)
        if self.disable: return self

        # update progress bar
        if kw:
            M.pbar.update(self.task_id, **kw)
        return self


def ipool(
        fn: Callable, 
        iter: Iterable, 
        desc: str='', 
        bar_kw: dict={}, 
        n_workers: int=8, 
        pool_mode: T_POOL_MODE='process', 
        ordered_: bool=False, 
        squeeze_: bool=False,
        **kw) -> Iterable:
    """Execute a function in a process pool with a progress bar for each task.

    Args:
        fn (Callable): The function to execute.
        iter (Iterable): The iterable to iterate over.
        desc (str, optional): The description of the main progress bar. 
        bar_kw (dict, optional): Additional keyword arguments for the sub progress bars.
        n_workers (int, optional): The number of workers in the pool. Defaults to 8.
        pool_mode (str, optional): The mode of the pool. Can be 'process', 'thread', 'sequential'. Defaults to 'process'.
        ordered_ (bool, optional): Whether to yield the results in order. Defaults to False.
        squeeze_ (bool, optional): Whether to skip the pool and main progress bar if there is only one item in the iterable. Defaults to False.
    """
    # no workers, just run sequentially
    if n_workers in {0, 1}:
        pool_mode = 'sequential'

    # if the iterable is a single item, just run the function
    if squeeze_ and utils.try_len(iter, -1) == 1:
        arg = utils.args.from_item(iter[0], **kw)
        yield arg(fn)
        return

    bar_kw = {'transient': True, **(bar_kw or {})}
    try:
        # initialize progress bars and run the tasks in a process/thread pool
        with mqdm(desc=desc, **bar_kw) as pbars:
            with M.executor(pool_mode, bar_kw=bar_kw, max_workers=n_workers) as executor:
                futures = []
                for arg in iter:
                    arg = utils.args.from_item(arg, **kw)
                    futures.append(executor.submit(fn, *arg.a, **arg.kw))
                    pbars.set(append_total=1)

                # get function results
                for f in futures if ordered_ else as_completed(futures):
                    x = f.result()
                    pbars.update()
                    yield x
    except:
        # pause the progress bar so it doesn't interfere with the traceback
        M.pause()
        raise


@wraps(ipool, ['__annotations__', '__doc__', '__type_params__'])
def pool(*a, results_: list=None, **kw) -> list:
    results_ = [] if results_ is None else results_
    results_.extend(ipool(*a, **kw))
    return results_
